/**
 * 响应式系统模块
 */

let activeFn = null
const targetMap = new WeakMap()

class Depend {
  constructor() {
    this.activeFns = new Set()
  }
  depend(){
    if(activeFn) this.activeFns.add(activeFn)
  }
  notify(){
    this.activeFns.forEach(fn => fn())
  }
}

function getDepend(target, key) {
  let depMap = targetMap.get(target)
  if(!depMap) {
    depMap = new Map()
    targetMap.set(target, depMap)
  }
  let dep = depMap.get(key)
  if(!dep) {
    dep = new Depend()
    depMap.set(key, dep)
  }
  return dep
}

/* 基于Proxy实现(Vue3) */
function reactive(obj, deep = true) {
  return new Proxy(obj, {
    get: function (target, key, receiver) {
      const dep = getDepend(target, key)
      dep.depend()
      const res = Reflect.get(target, key, receiver)
      if(!deep) return res
      /* deep: true -> 递归调用 */
      if(isObject(res)) return reactive(res)
      return res
    },
    set: function (target, key, value, receiver) {
      Reflect.set(target, key, value, receiver)
      const dep = getDepend(target, key)
      dep.notify()
      return true
    }
  })
}

/* 基于Object.defineProperty实现(Vue2) */
// function reactive(obj) {
//   Object.keys(obj).forEach(key => {
//     let value = obj[key]
//     Object.defineProperty(obj, key, {
//       get: function () {
//         const dep = getDepend(obj, key)
//         dep.depend()
//         return value
//       },
//       set: function (newValue) {
//         value = newValue
//         const dep = getDepend(obj, key)
//         dep.notify()
//       }
//     })
//   })
//   return obj
// }

function isObject(value) {
  const type = typeof value
  return (type === 'object' || type === 'function') && value !== null
}

function watchEffect(fn) {
  activeFn = fn
  fn()
  activeFn = null
}

export {
  reactive,
  watchEffect
}

/**
 * test demo
 */

 const data = reactive({
  name: "Michael",
  age: 18,
  friends: [{
    name: "Sara",
    age: 19
  }]
})

watchEffect(() => console.log(data.name))
watchEffect(() => console.log(data.age))
watchEffect(() => console.log(data.friends[0].name))
watchEffect(() => console.log(data.friends[0].age))

data.name = "Zhiping Li"
data.age++
data.friends[0].name = "Rachael"
data.friends[0].age += 2
